import copy
import json
import time
from typing import (
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple
)

from dataladmetadatamodel import (
    JSONObject,
    check_serialized_version,
    version_string
)
from dataladmetadatamodel.mappableobject import MappableObject
from dataladmetadatamodel.mapper.reference import Reference


class ParameterDict(dict):
    def __hash__(self):
        return hash(tuple(sorted(self.items())))


class ExtractorConfiguration:
    """
    Holds a single configuration for an extractor.
    Ensures that a version number is at least given.
    """
    def __init__(self,
                 version: str,
                 parameter: Dict[str, JSONObject]):
        self.version = version
        self.parameter = ParameterDict(parameter)

    def to_json_obj(self) -> JSONObject:
        return {
            "@": dict(
                type="ExtractorConfiguration",
                version=version_string
            ),
            "version": self.version,
            "parameter": self.parameter
        }

    def to_json_str(self) -> str:
        return json.dumps(self.to_json_obj())

    def __eq__(self, other):
        return (
            self.version == other.version
            and self.parameter == other.parameter
        )

    def __hash__(self):
        return hash((self.version, self.parameter))

    @classmethod
    def from_json_obj(cls, obj: JSONObject) -> "ExtractorConfiguration":
        assert obj["@"]["type"] == "ExtractorConfiguration"
        check_serialized_version(obj)
        return cls(
            obj["version"],
            obj["parameter"]
        )

    @classmethod
    def from_json_str(cls, json_str: str) -> "ExtractorConfiguration":
        return cls.from_json_obj(json.loads(json_str))


class MetadataInstance:
    """
    A single metadata instance. It is associated
    with provenance information, i.e. time stamp,
    author, author_email, with a configuration, i.e.
    parameters, and with a source that points to
    the metadata itself.
    """
    def __init__(self,
                 time_stamp,
                 author_name,
                 author_email,
                 configuration: ExtractorConfiguration,
                 metadata_content: JSONObject):

        self.time_stamp = time_stamp
        self.author_name = author_name
        self.author_email = author_email
        self.configuration = configuration
        self.metadata_content = metadata_content

    def to_json_obj(self) -> JSONObject:
        return {
            "@": dict(
                type="MetadataInstance",
                version=version_string
            ),
            "time_stamp": self.time_stamp,
            "author": self.author_name,
            "author_email": self.author_email,
            "configuration": self.configuration.to_json_obj(),
            "metadata_content": self.metadata_content
        }

    def to_json_str(self) -> str:
        return json.dumps(self.to_json_obj())

    def __eq__(self, other):
        return (
                self.time_stamp == other.time_stamp
                and self.author_name == other.author_name
                and self.author_email == other.author_email
                and self.configuration == other.configuration
                and self.metadata_content == other.metadata_content
        )

    @classmethod
    def from_json_obj(cls, obj: JSONObject) -> "MetadataInstance":
        assert obj["@"]["type"] == "MetadataInstance"
        check_serialized_version(obj)
        return cls(
            obj["time_stamp"],
            obj["author"],
            obj["author_email"],
            ExtractorConfiguration.from_json_obj(obj["configuration"]),
            obj["metadata_content"]
        )

    @classmethod
    def from_json_str(cls, json_str: str) -> "MetadataInstance":
        return cls.from_json_obj(json.loads(json_str))


class MetadataInstanceSet:
    """
    A set of metadata instances, i.e. extractor
    run information records. Each instance is
    identified by its configuration, i.e. an
    instance of ExtractorConfiguration.
    """
    def __init__(self,
                 initial_metadata_instances: Optional[Iterable[MetadataInstance]] = None):

        self.parameter_set = list()
        self._instances = dict()
        for metadata_instance in initial_metadata_instances or []:
            self.add_metadata_instance(metadata_instance)

    def __iter__(self):
        yield from self._instances.values()

    def add_metadata_instance(self, metadata_instance: MetadataInstance):
        if metadata_instance.configuration not in self.parameter_set:
            self.parameter_set.append(metadata_instance.configuration)
        instance_key = self.parameter_set.index(metadata_instance.configuration)
        self._instances[instance_key] = metadata_instance

    @property
    def instances(self) -> Dict:
        return self._instances

    @property
    def configurations(self) -> List[ExtractorConfiguration]:
        return self.parameter_set[:]

    def get_instance_for_configuration_index(self, index: int):
        return self._instances[index]

    def get_instance_for_configuration(self, configuration: ExtractorConfiguration):
        return self._instances[self.parameter_set.index(configuration)]

    def to_json_obj(self) -> JSONObject:
        return {
            "@": dict(
                type="MetadataInstanceSet",
                version=version_string
            ),
            "parameter_set": [
                configuration.to_json_obj()
                for configuration in self.parameter_set
            ],
            "instance_set": {
                instance_key: instance.to_json_obj()
                for instance_key, instance in self._instances.items()
            }
        }

    def to_json_str(self) -> str:
        return json.dumps(self.to_json_obj())

    @classmethod
    def from_json_obj(cls, obj: JSONObject) -> "MetadataInstanceSet":
        assert obj["@"]["type"] == "MetadataInstanceSet"
        check_serialized_version(obj)

        metadata_instance_set = cls()
        metadata_instance_set.parameter_set = [
            ExtractorConfiguration.from_json_obj(json_obj)
            for json_obj in obj["parameter_set"]
        ]
        metadata_instance_set._instances = {
            int(configuration_id): MetadataInstance.from_json_obj(json_obj)
            for configuration_id, json_obj in obj["instance_set"].items()
        }
        return metadata_instance_set

    @classmethod
    def from_json_str(cls, json_str: str) -> "MetadataInstanceSet":
        return cls.from_json_obj(json.loads(json_str))

    def __eq__(self, other: "MetadataInstanceSet"):
        return sorted(self.parameter_set) == sorted(other.parameter_set) \
               and self._instances == other._instances


class Metadata(MappableObject):
    """
    Holds entries for all metadata of a single object.
    Metadata is identified on the first level by its
    format-name, i.e. the extractor-name. For each
    extractor there is a set of configurations and
    associated metadata, i.e. objects that contain
    the extractor result, aka the real metadata.
    """
    def __init__(self,
                 realm: Optional[str] = None,
                 reference: Optional[Reference] = None):

        assert isinstance(realm, (type(None), str))
        assert isinstance(reference, (type(None), Reference))

        super().__init__(realm, reference)
        self.instance_sets: Dict[str, MetadataInstanceSet] = dict()

    def __eq__(self, other):
        return self.instance_sets == other.instance_sets

    @staticmethod
    def get_empty_instance(realm: Optional[str] = None,
                           reference: Optional[Reference] = None):
        return Metadata(realm, reference)

    def purge_impl(self):
        self.instance_sets = dict()

    def modifiable_sub_objects_impl(self) -> Iterable[MappableObject]:
        return []

    @property
    def extractors(self) -> Generator[str, None, None]:
        self.ensure_mapped()
        yield from self.instance_sets.keys()

    @property
    def extractor_runs(self) -> Generator[Tuple[str, MetadataInstanceSet], None, None]:
        self.ensure_mapped()
        yield from self.instance_sets.items()

    def extractor_runs_for_extractor(self, extractor_name: str) -> MetadataInstanceSet:
        self.ensure_mapped()
        return self.instance_sets[extractor_name]

    def add_extractor_run(self,
                          time_stamp: Optional[float],
                          extractor_name: str,
                          author_name: str,
                          author_email: str,
                          configuration: ExtractorConfiguration,
                          metadata_content: JSONObject):

        self.ensure_mapped()
        self.touch()

        instance_set = self.instance_sets.get(
            extractor_name,
            MetadataInstanceSet())

        instance_set.add_metadata_instance(
            MetadataInstance(
                time_stamp if time_stamp is not None else time.time(),
                author_name,
                author_email,
                configuration,
                metadata_content))

        self.instance_sets[extractor_name] = instance_set

    def to_json(self) -> str:
        return json.dumps({
            "@": dict(
                type="Metadata",
                version=version_string
            ),
            "instance_sets": {
                format_name: instance_set.to_json_obj()
                for format_name, instance_set in self.instance_sets.items()
            }
        })

    def init_from_json(self, json_str) -> None:
        obj = json.loads(json_str)
        check_serialized_version(obj)

        assert obj["@"]["type"] == "Metadata"
        for format_name, instance_set_json_obj in obj["instance_sets"].items():
            self.instance_sets[format_name] = \
                MetadataInstanceSet.from_json_obj(instance_set_json_obj)

    @classmethod
    def from_json(cls, json_str: str) -> "Metadata":
        metadata = cls(None)
        metadata.init_from_json(json_str)
        metadata.mapped = True
        metadata.set_unsaved()
        return metadata

    def deepcopy_impl(self,
                      new_mapper_family: Optional[str] = None,
                      new_destination: Optional[str] = None,
                      **kwargs) -> "Metadata":

        copied_metadata = Metadata()
        for extractor_name, instance_set in self.instance_sets.items():
            copied_metadata.instance_sets[extractor_name] = copy.deepcopy(instance_set)
        copied_metadata.write_out(new_destination)
        copied_metadata.purge()
        return copied_metadata



import json
import unittest
from pathlib import Path
from typing import cast
from unittest import mock

from .... import version_string
from ....tests.utils import get_location
from dataladmetadatamodel.mapper import get_mapper
from dataladmetadatamodel.mapper.gitmapper.metadatamapper import MetadataGitMapper
from dataladmetadatamodel.mapper.gitmapper.objectreference import flush_object_references
from dataladmetadatamodel.metadata import (
    ExtractorConfiguration,
    Metadata
)


expected_reference_object = {
    "@": {
        "type": "Reference",
        "version": version_string
    },
    "class_name": "Metadata",
    "location": get_location(1)
}


expected_metadata_object = {
    "@": {
        "type": "Metadata",
        "version": version_string
    },
    "instance_sets": {
        "test-extractor": {
            "@": {
                "type": "MetadataInstanceSet",
                "version": version_string
            },
            "parameter_set": [
                {
                    "@": {
                        "type": "ExtractorConfiguration",
                        "version": version_string
                    },
                    "version": "v3.4",
                    "parameter": {"p1": "1"}
                }
            ],
            "instance_set": {
                "0": {
                    "@": {
                        "type": "MetadataInstance",
                        "version": version_string
                    },
                    "time_stamp": 1.2,
                    "author": "test-tool",
                    "author_email": "test-tool@test.com",
                    "configuration": {
                        "@": {
                            "type": "ExtractorConfiguration",
                            "version": version_string
                        },
                        "version": "v3.4",
                        "parameter": {"p1": "1"}
                    },
                    "metadata_content": {
                        "key1": "this is metadata"
                    }
                }
            }
        }
    }
}


class TestMetadataMapper(unittest.TestCase):

    def test_basic_unmapping(self):
        metadata = Metadata()
        metadata.add_extractor_run(
            1.2,
            "test-extractor",
            "test-tool",
            "test-tool@test.com",
            ExtractorConfiguration("v3.4", {"p1": "1"}),
            {"key1": "this is metadata"})

        with \
                mock.patch("dataladmetadatamodel.mapper.gitmapper"
                           ".metadatamapper.git_save_str") as save_str, \
                mock.patch("dataladmetadatamodel.mapper.gitmapper"
                           ".treeupdater.git_save_tree_node") as save_tree_node, \
                mock.patch("dataladmetadatamodel.mapper.gitmapper"
                           ".gitbackend.subprocess.git_update_ref") as update_ref:

            save_str.return_value = get_location(1)
            save_tree_node.return_value = get_location(2)
            update_ref.return_value = get_location(3)

            realm = "/tmp/t1"
            reference = metadata.write_out(realm, "git")
            flush_object_references(Path("/tmp/t1"))

            self.assertEqual(len(save_str.call_args_list), 2)
            self.assertEqual(
                json.loads(save_str.call_args_list[0][0][1]),
                expected_metadata_object)
            self.assertEqual(
                json.loads(save_str.call_args_list[1][0][1]),
                expected_reference_object)

    def test_double_cache_detection(self):
        metadata_mapper: MetadataGitMapper = cast(
            MetadataGitMapper,
            get_mapper("Metadata", "git")
        )
        metadata_mapper.cache_realm("a")
        self.assertRaises(RuntimeError, metadata_mapper.cache_realm, "a")
        self.assertRaises(RuntimeError, metadata_mapper.flush_realm, "err")


if __name__ == '__main__':
    unittest.main()



from abc import (
    ABCMeta,
    abstractmethod
)
from typing import (
    Iterable,
    Optional,
)


class ModifiableObject(metaclass=ABCMeta):
    """
    Modifiable objects track their modification status.
    including the modification status of potential
    contained objects.

    If one of the contained objects is modified, the
    containing object is considered modified as well.

    Objects determine their modification-relevant
    sub-objects by implementing their version of the
    method "get_modifiable_sub_objects".
    """
    def __init__(self, saved_on: Optional[str] = None):
        self.saved_on = set()
        if saved_on:
            self.saved_on.add(saved_on)

    def touch(self):
        self.set_unsaved()

    def set_saved_on(self, destination: str):
        self.saved_on.add(destination)

    def set_unsaved(self):
        self.saved_on = set()

    def is_saved_on(self, destination: str) -> bool:
        """
        Check whether the object is saved on the given
        destination. This check includes whether all
        modifiable sub-objects are saved on the
        given destination.
        """
        if destination not in self.saved_on:
            return False
        return all(
            map(
                lambda element: element.is_saved_on(destination),
                self.modifiable_sub_objects))

    @abstractmethod
    def modifiable_sub_objects(self) -> Iterable["ModifiableObject"]:
        raise NotImplementedError



from abc import (
    ABCMeta,
    abstractmethod
)
from typing import Optional

from dataladmetadatamodel.log import logger
from dataladmetadatamodel.mapper.reference import Reference


class Mapper(metaclass=ABCMeta):
    """
    Mapper are responsible for populating an existing
    object of a certain class from a backend. The mapper
    will delegate sub-object mapping to the mapper for
    the class of the sub-object.

    This base class handles default destination
    """
    def __init__(self,
                 class_name: str,
                 default_destination: Optional[str] = None):
        self.class_name = class_name
        self.destination = default_destination

    def map_in(self,
               mappable_object: "MappableObject",
               realm: str,
               reference: Reference) -> None:

        # Check for old-style file tree and dataset tree references
        if reference.class_name in ("FileTree", "DatasetTree"):
            logger.warning(f"Mapper.map_in(): converting reference to class "
                           f"{reference.class_name} to a reference to class "
                           f"MTreeNode")
            reference.class_name = "MTreeNode"

        # TODO: this is not too nice, but required since FileTree and
        #  DatasetTree are proxies for MTreeNode which just add some
        #  methods.
        mo_class_name = type(mappable_object).__name__
        if mo_class_name in ("FileTree", "DatasetTree"):
            assert self.class_name == "MTreeNode", \
                f"Mappable object class: {mo_class_name} " \
                f"can not be mapped in by a mapper for class: {self.class_name}"
        else:
            assert mo_class_name == self.class_name, \
                f"Mappable object class name ({mo_class_name}) " \
                f"does not match this mapper's class_name ({self.class_name})"

        assert reference.class_name == self.class_name, \
            f"Reference class name ({reference.class_name}) " \
            f"does not match self.class_name ({self.class_name})"

        self.map_in_impl(mappable_object, realm, reference)

    def map_out(self,
                mappable_object: "MappableObject",
                realm: Optional[str] = None,
                force_write: bool = False) -> Reference:

        if realm is None:
            if self.destination is None:
                raise Exception(
                    "'destination' not set and no default destination provided")
            realm = self.destination
        return self.map_out_impl(mappable_object, realm, force_write)

    @abstractmethod
    def map_in_impl(self,
                    mappable_object: "MappableObject",
                    realm: str,
                    reference: Reference) -> None:
        raise NotImplementedError

    @abstractmethod
    def map_out_impl(self,
                     mappable_object: "MappableObject",
                     realm: str,
                     force_write: bool) -> Reference:
        raise NotImplementedError



import hashlib
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import (
    Dict,
    List,
    Tuple,
    Union,
)

from .gitbackend.subprocess import git_save_file_list


def hash_blob(blob: Union[str, bytes]) -> str:
    if isinstance(blob, str):
        blob = bytearray(blob.encode())
    else:
        blob = bytearray(blob)

    object_to_hash = f"blob {len(blob)}".encode() + b"\x00" + blob
    return hashlib.sha1(object_to_hash).hexdigest()


class GitBlobCache:
    def __init__(self,
                 realm: str,
                 maxsize: int = 2000):

        self.realm = realm
        self.maxsize = maxsize
        self.cached_objects: List[Tuple[Union[str, bytes], str]] = list()
        self.flushed_objects: Dict[Union[str, bytes], str] = dict()
        self.temporary_directory = TemporaryDirectory()
        self.temp_dir_path = Path(self.temporary_directory.name)

        # Assert after member initialisation, so in case of an
        # exception the destructor does still work
        assert isinstance(realm, str)

    def __del__(self):
        if len(self.cached_objects) > 0:
            raise RuntimeError("deleting a non-flushed JSON object cache")
        self.temporary_directory.cleanup()

    def cache_blob(self,
                   realm: str,
                   blob: Union[str, bytes]):

        assert realm == self.realm, \
            "realm of cached object and realm of cache instance differ"

        if len(self.cached_objects) == self.maxsize:
            self.flush()
        expected_hash = hash_blob(blob)
        self.cached_objects.append((blob, expected_hash))
        return expected_hash

    def flush(self):
        """
        Write all cached objects to a git repository,
        associate the objects with their hash in
        self.cached_objects.

        Writing is done by creating files in a
        temporary directory and calling git hash-object
        with the list of files.

        :return: None
        """
        def check_hash(hash: str, expected_hash: str):
            assert hash == expected_hash

        file_list = []
        for index, (blob, expected_hash) in enumerate(self.cached_objects):
            temp_file_path = self.temp_dir_path / str(index)
            with temp_file_path.open("tw") as f:
                f.write(blob)
            file_list.append(str(temp_file_path))

        hash_values = git_save_file_list(self.realm, file_list)
        assert len(hash_values) == len(file_list), \
            f"hash value list length ({len(hash_values)}) and file list length " \
            f"({len(file_list)}) differ.\n{hash_values}\n{file_list}"

        hash_dict = {
            blob: hash_values[index]
            for index, (blob, expected_hash) in enumerate(self.cached_objects)
            if check_hash(hash_values[index], expected_hash)
        }
        self.flushed_objects.update(hash_dict)
        self.cached_objects = []



from typing import (
    cast,
    List,
    Optional,
    Tuple,
)

from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord
from dataladmetadatamodel.mtreenode import MTreeNode
from dataladmetadatamodel.mtreeproxy import MTreeProxy
from dataladmetadatamodel.mapper.reference import Reference


datalad_root_record_name = ".datalad_metadata_root_record"


class DatasetTree(MTreeProxy):
    def __init__(self,
                 mtree: Optional[MTreeNode] = None,
                 realm: Optional[str] = None,
                 reference: Optional[Reference] = None):

        assert isinstance(mtree, (type(None), MTreeNode))
        assert isinstance(realm, (type(None), str))
        assert isinstance(reference, (type(None), Reference))

        super().__init__(MetadataRootRecord, mtree, realm, reference)

    def __contains__(self, path: MetadataPath) -> bool:
        return self.mtree.get_object_at_path(path / datalad_root_record_name) is not None

    def add_dataset(self,
                    path: MetadataPath,
                    metadata_root_record: MetadataRootRecord):

        self.mtree.add_child_at(metadata_root_record,
                                path / datalad_root_record_name)

    def get_metadata_root_record(self,
                                 path: MetadataPath
                                 ) -> Optional[MetadataRootRecord]:

        mrr = self.mtree.get_object_at_path(
            path / datalad_root_record_name)

        if mrr is None:
            return None

        assert isinstance(mrr, MetadataRootRecord)
        mrr.ensure_mapped()
        return mrr

    @property
    def dataset_paths(self
                      ) -> List[Tuple[MetadataPath, MetadataRootRecord]]:
        return [
            (
                MetadataPath("/".join(path.parts[:-1])),
                cast(MetadataRootRecord, node)
            )
            for path, node in self.mtree.get_paths_recursive()
            if path.parts[-1] == datalad_root_record_name
        ]

    def add_subtree(self,
                    subtree: "DatasetTree",
                    subtree_path: MetadataPath):

        self.mtree.add_child_at(subtree.mtree, subtree_path)

    def delete_subtree(self,
                       subtree_path: MetadataPath):
        self.mtree.remove_child_at(subtree_path)



from typing import (
    Any,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from dataladmetadatamodel.mappableobject import (
    MappableObject,
    ensure_mapped,
)
from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.mapper.reference import Reference


class MTreeNode(MappableObject):
    def __init__(self,
                 leaf_class: Any,
                 realm: Optional[str] = None,
                 reference: Optional[Reference] = None):

        assert isinstance(reference, (type(None), Reference))

        super().__init__(realm, reference)
        self.leaf_class = leaf_class
        self.leaf_class_name = leaf_class.__name__
        self.child_nodes = dict()

    def __contains__(self, path: MetadataPath) -> bool:
        return self.get_child(path) is not None

    def __str__(self):
        return f"<{type(self).__name__}: children: {self.child_nodes.keys()}>"

    def modifiable_sub_objects_impl(self) -> Iterable["MappableObject"]:
        yield from self.child_nodes.values()

    def purge_impl(self):
        for child_node in self.child_nodes.values():
            child_node.purge()
        self.child_nodes = dict()

    def deepcopy_impl(self,
                      new_mapper_family: Optional[str] = None,
                      new_destination: Optional[str] = None,
                      **kwargs) -> "MappableObject":

        copied_mtree_node = MTreeNode(self.leaf_class)
        for child_name, child_node in self.child_nodes.items():
            copied_mtree_node.add_child(
                child_name,
                child_node.deepcopy(
                    new_mapper_family,
                    new_destination))

        copied_mtree_node.write_out(new_destination)
        copied_mtree_node.purge()
        return copied_mtree_node

    def contains_child(self,
                       child_name: Union[str, MetadataPath]
                       ) -> bool:

        return self.get_child(child_name) is not None

    def get_child(self,
                  child_name: Union[str, MetadataPath]
                  ) -> Optional[MappableObject]:

        self.ensure_mapped()
        return self.child_nodes.get(str(child_name), None)

    def add_child(self, node_name: str, child: MappableObject):
        """
        Add a sub-element with the given name. The sub-element
        can be an MTreeNode or any other MappableObject (usually
        it will be an MTreeNode, Metadata, or MetadataRootRecord.
        """
        self._add_children([(node_name, child)])

    def _add_children(self,
                      children: List[Tuple[str, MappableObject]]):

        # Assert that no name is "." or contains a "/", and
        # assert that there are no duplicated names in the input
        assert all(map(lambda nn: nn[0] != ".", children))
        assert all(map(lambda nn: nn[0].find("/") < 0, children))
        assert len(set(map(lambda nn: nn[0], children))) == len(children)

        new_names = set(map(lambda nn: nn[0], children))

        self.ensure_mapped()
        duplicated_names = set(self.child_nodes.keys()) & new_names
        if duplicated_names:
            raise ValueError(
                "Child node with the following name(s) already exist(s): "
                + ", ".join(duplicated_names))

        self.touch()
        self.child_nodes = {
            **self.child_nodes,
            **dict(children)
        }

    def add_child_at(self,
                     child: MappableObject,
                     path: MetadataPath):

        assert len(path) > 0, "child name required"

        path_element = path.parts[0]
        existing_child = self.get_child(path_element)

        if len(path) == 1:
            if isinstance(existing_child, MTreeNode):
                raise ValueError(
                    f"tree-node named {path_element} exists: "
                    f"{existing_child}, cannot be converted "
                    f"to {child}")

            # Add the final object to self
            self.add_child(path.parts[0], child)

        else:

            if existing_child is None:
                # Create an intermediate node, if it doesn't exist
                existing_child = MTreeNode(self.leaf_class)
                self.add_child(path_element, existing_child)

            if not isinstance(existing_child, MTreeNode):
                raise ValueError(
                    f"non tree-node named {path_element} exists: "
                    f"{existing_child}, cannot be converted to a "
                    f"tree-node")

            existing_child.add_child_at(
                child,
                MetadataPath(
                    "/".join(path.parts[1:])))

    def remove_child(self,
                     child_name: str):

        self.ensure_mapped()
        self.touch()
        del self.child_nodes[child_name]

    def remove_child_at(self,
                        path: MetadataPath):

        containing_path = MetadataPath("/".join(path.parts[:-1]))
        containing_node = self.get_object_at_path(containing_path)
        containing_node.remove_child(path.parts[-1])

    def get_object_at_path(self,
                           path: Optional[MetadataPath] = None
                           ) -> Optional["MTreeNode"]:

        # Linear search for prefix_path
        path = path or MetadataPath("")
        current_node = self
        for element in path.parts:
            if not isinstance(current_node, MTreeNode):
                return None
            current_node = current_node.get_child(element)
            if current_node is None:
                return None
        return current_node

    def get_paths_recursive(self,
                            show_intermediate: Optional[bool] = False
                            ) -> Iterable[Tuple[MetadataPath, "MappableObject"]]:

        if show_intermediate:
            yield MetadataPath(""), self

        with ensure_mapped(self):
            for child_name, child_node in self.child_nodes.items():
                if not isinstance(child_node, MTreeNode):
                    yield MetadataPath(child_name), child_node
                else:
                    for sub_path, tree_node in child_node.get_paths_recursive(
                            show_intermediate):
                        yield MetadataPath(child_name) / sub_path, tree_node

    @staticmethod
    def is_root_path(path: MetadataPath) -> bool:
        return len(path.parts) == 0




from dataladmetadatamodel.mapper.gitmapper.gitbackend.subprocess import (
    git_read_tree_node,
    git_save_tree_node,
)
from dataladmetadatamodel.mapper.mapper import Mapper
from dataladmetadatamodel.mapper.reference import Reference


class MTreeNodeGitMapper(Mapper):

    def map_in_impl(self,
                    mtree_node: "MTreeNode",
                    realm: str,
                    reference: Reference) -> None:

        from dataladmetadatamodel.mtreenode import MTreeNode

        if reference.is_none_reference():
            mtree_node.child_nodes = dict()
            return

        assert isinstance(mtree_node, MTreeNode)

        lines = git_read_tree_node(realm,
                                   reference.location)

        for entry in [line.split() for line in lines]:
            if entry[1] == "tree":
                child = MTreeNode(
                    leaf_class=mtree_node.leaf_class,
                    realm=realm,
                    reference=Reference("MTreeNode",
                                        entry[2]))

            elif entry[1] == "blob":
                child = mtree_node.leaf_class.get_empty_instance(
                    realm=realm,
                    reference=Reference(mtree_node.leaf_class_name,
                                        entry[2]))

            else:
                raise ValueError(f"unknown git tree entry type: {entry[1]}")
            mtree_node.child_nodes[entry[3]] = child

    def map_out_impl(self,
                     mtree_node: "MTreeNode",
                     realm: str,
                     force_write: bool) -> Reference:

        from dataladmetadatamodel.mtreenode import MTreeNode

        assert isinstance(mtree_node, MTreeNode)

        if not mtree_node.child_nodes:
            return Reference.get_none_reference("MTreeNode")

        for child_node in mtree_node.child_nodes.values():
            child_node.write_out(realm)

        dir_entries = [
            (
                "040000" if isinstance(child_node, MTreeNode) else "100644",
                "tree" if isinstance(child_node, MTreeNode) else "blob",
                child_node.reference.location,
                child_name
            )
            for child_name, child_node in mtree_node.child_nodes.items()
        ]
        mtree_node_location = git_save_tree_node(realm, dir_entries)
        return Reference("MTreeNode", mtree_node_location)



import logging
import time
from typing import Dict, Optional, Tuple
from uuid import UUID

from dataladmetadatamodel.metadata import ExtractorConfiguration, Metadata
from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord

from tools.metadata_creator.filetreecreator import create_file_tree
from tools.metadata_creator.utils import get_dataset_id, get_dataset_version, \
    read_datasets


mdc_logger = logging.getLogger("metadata_creator")


def create_metadata_root_record(mapper_family,
                                realm,
                                dataset_id: UUID,
                                dataset_version: str,
                                dataset_path: str,
                                relative_path: str,
                                parameter_set_count: int
                                ) -> Optional[MetadataRootRecord]:

    file_tree = create_file_tree(
        mapper_family,
        realm,
        dataset_path,
        parameter_set_count
    )

    metadata = Metadata()
    for count in range(parameter_set_count):
        parameters = {
            "ds_parameter_0": f"value_0.{count}",
            "ds_parameter_1": f"value_1.{count}",
        }
        metadata.add_extractor_run(
                time.time(),
                "dataset-core-extractor",
                "datasetcreator.py",
                "support@datalad.org",
                ExtractorConfiguration("1.2.3", parameters),
                {
                    "info": f"dataset-level test metadata for parameter set #{count}",
                    "prefix_path": relative_path
                }
            )

    mrr = MetadataRootRecord(
        mapper_family,
        realm,
        dataset_id,
        dataset_version,
        Connector.from_object(metadata),
        Connector.from_object(file_tree)
    )

    mrr.save()
    mrr.dataset_level_metadata.purge()
    mrr._file_tree.purge()
    return mrr


def get_dataset_id_version(path: str) -> Tuple[Optional[UUID], Optional[str]]:

    dataset_id = get_dataset_id(path)
    if dataset_id is None:
        mdc_logger.error(f"cannot determine id of dataset at {path}")

    dataset_version = get_dataset_version(path)
    if dataset_version is None:
        mdc_logger.error(f"cannot determine version of dataset at {path}")

    return dataset_id, dataset_version


def create_mrrs_from_dataset(mapper: str,
                             realm: str,
                             root_path: str,
                             parameter_set_count: int
                             ) -> Dict[Tuple[UUID, str, str], MetadataRootRecord]:

    result = dict()
    for relative_path, entry in read_datasets(root_path):

        dataset_path = entry.path

        dataset_id, dataset_version = get_dataset_id_version(dataset_path)
        if dataset_id is None or dataset_version is None:
            mdc_logger.info(f"ignoring dataset at {dataset_version} because version or id could not be read")
            continue

        mrr = create_metadata_root_record(
            mapper,
            realm,
            dataset_id,
            dataset_version,
            dataset_path,
            relative_path,
            parameter_set_count)

        result[(dataset_id, dataset_version, relative_path)] = mrr

    return result



import unittest
from typing import (
    Iterable,
    Optional
)
from unittest.mock import MagicMock

from dataladmetadatamodel.mappableobject import MappableObject
from dataladmetadatamodel.mapper import set_mapper
from dataladmetadatamodel.mapper.reference import Reference


class SUTMappableObject(MappableObject):
    def __init__(self,
                 realm=None,
                 reference=None):

        super().__init__(realm, reference)
        self.something = "Something " * 100

    def deepcopy_impl(self,
                      new_mapper_family: Optional[str] = None,
                      new_destination: Optional[str] = None,
                      **kwargs) -> "MappableObject":
        return self

    def purge_impl(self):
        self.something = None

    def modifiable_sub_objects_impl(self) -> Iterable:
        return []


class TestMappableObject(unittest.TestCase):

    def setUp(self) -> None:
        self.test_mapper = MagicMock(
            map_out=MagicMock(return_value=Reference(
                "SUTMappableObject",
                "location:00001111"
            ))
        )
        set_mapper("SUTMappableObject", "git", self.test_mapper)

    def test_new_mapped(self):
        # expect that a newly created object without reference is written out once
        mo = SUTMappableObject()
        mo.write_out("/tmp/t", "git")
        self.test_mapper.map_out.assert_called_once_with(mo, "/tmp/t", False)
        mo.write_out("/tmp/t", "git")
        self.test_mapper.map_out.assert_called_once_with(mo, "/tmp/t", False)

    def test_purge(self):
        mo = SUTMappableObject(None)
        self.assertRaises(ValueError, mo.purge)
        mo.write_out("/tmp/t", "git")
        mo.purge()
        self.test_mapper.map_out.assert_called_once_with(mo, "/tmp/t", False)

    def test_purge_unsaved(self):
        mo = SUTMappableObject(None)
        self.assertRaises(ValueError, mo.purge)


if __name__ == '__main__':
    unittest.main()



import subprocess
import tempfile
import unittest
from typing import List
from unittest import mock

from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.mtreenode import MTreeNode
from dataladmetadatamodel.text import Text
from dataladmetadatamodel.mapper.reference import Reference


tree_location = "a000000000000000000000000000000000000000"
file_location = "a001000000000000000000000000000000000001"


def create_tree(file_names: List[str],
                sub_dir_names: List[str]
                ) -> MTreeNode:

    root_node = MTreeNode(Text)
    for sub_dir_name in sub_dir_names:
        sub_tree_node = MTreeNode(Text)
        for file_name in file_names:
            sub_tree_node.add_child(
                file_name,
                Text(f"content of: /{sub_dir_name}/{file_name}"))
        root_node.add_child(sub_dir_name, sub_tree_node)
    return root_node


class TestMTreeNodeMapper(unittest.TestCase):

    def test_basic_unmapping(self):

        realm = "/tmp/t1"
        file_names = ["a", "b", "c"]
        sub_dir_names = ["sub0", "sub1", "sub2"]

        root_node = create_tree(file_names, sub_dir_names)

        with \
                mock.patch(
                    "dataladmetadatamodel.mapper.gitmapper"
                    ".mtreenodemapper.git_save_tree_node") as save_tree_node, \
                mock.patch(
                    "dataladmetadatamodel.mapper.gitmapper" 
                    ".textmapper.git_save_str") as save_str:

            save_tree_node.configure_mock(return_value=tree_location)
            save_str.configure_mock(return_value=file_location)

            root_node.write_out(realm, "git")

            # Expect one (root) plus three (sub-dir trees) calls
            self.assertEqual(save_tree_node.call_count, 4)

            expected_sub_tree_calls = [
                mock.call(
                    realm,
                    [
                        ("100644", "blob", file_location, file_name)
                        for file_name in file_names
                    ]
                )
                for _ in range(len(sub_dir_names))
            ]
            save_tree_node.assert_has_calls(expected_sub_tree_calls, any_order=True)

            expected_root_calls = [
                mock.call(
                    realm,
                    [
                        ("040000", "tree", tree_location, sub_dir_name)
                        for sub_dir_name in sub_dir_names
                    ]
                )
            ]
            save_tree_node.assert_has_calls(expected_root_calls, any_order=True)

    def test_none_mapping_out(self):
        tree = MTreeNode(Text)
        reference = tree.write_out("/tmp/t1")
        self.assertTrue(reference.is_none_reference())

    def test_none_mapping_in(self):
        reference = Reference.get_none_reference("MTreeNode")
        tree = MTreeNode(Text, "/tmp/t1", reference)
        tree.read_in()
        self.assertEqual(len(tree.child_nodes), 0)

    def test_mapping_end_to_end(self):
        file_names = ["a", "b", "c"]
        sub_dir_names = ["sub0", "sub1", "sub2"]

        with tempfile.TemporaryDirectory() as realm:

            subprocess.run(["git", "init", realm])

            root_node = create_tree(file_names, sub_dir_names)
            reference = root_node.write_out(realm)

            new_root_node = MTreeNode(
                leaf_class=Text,
                realm=realm,
                reference=reference)
            new_root_node.read_in()

            tree_elements = list(new_root_node.get_paths_recursive())
            self.assertEqual(len(tree_elements),
                             len(file_names) * len(sub_dir_names))
            self.assertEqual(
                [str(tree_element[0]) for tree_element in tree_elements],
                [
                    sub_dir_name + "/" + file_name
                    for sub_dir_name in sub_dir_names
                    for file_name in file_names
                ]
            )

            self.assertEqual(
                [
                    tree_element[1].read_in().content
                    for tree_element in tree_elements
                ],
                [
                    f"content of: /{sub_dir_name}/{file_name}"
                    for sub_dir_name in sub_dir_names
                    for file_name in file_names
                ]
            )


if __name__ == '__main__':
    unittest.main()




from dataladmetadatamodel.mapper.gitmapper.gitbackend.subprocess import (
    git_load_str,
    git_save_str
)
from dataladmetadatamodel.mapper.mapper import Mapper
from dataladmetadatamodel.mapper.reference import Reference


class TextGitMapper(Mapper):

    def map_in_impl(self,
                    text: "Text",
                    realm: str,
                    reference: Reference) -> None:

        from ...text import Text

        assert isinstance(text, Text)
        assert isinstance(reference, Reference)
        assert reference.class_name == "Text"

        text.content = git_load_str(realm, reference.location)
        # TODO: what about the references?

    def map_out_impl(self,
                     text: "Text",
                     realm: str,
                     force_write: bool) -> Reference:

        from ...text import Text

        assert isinstance(text, Text)

        text_location = git_save_str(realm, text.content)
        return Reference("Text", text_location)



import subprocess
import tempfile
import unittest

from dataladmetadatamodel.datasettree import DatasetTree
from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.uuidset import (
    UUIDSet,
    VersionList,
)
from dataladmetadatamodel.versionlist import VersionRecord

from .utils import (
    assert_uuid_sets_equal,
    create_dataset_tree,
    get_uuid
)


uuid_0 = get_uuid(0)
uuid_1 = get_uuid(1)
uuid_2 = get_uuid(2)


class TestUUIDSet(unittest.TestCase):

    def test_basic(self):
        UUIDSet()

    def test_copy_from_memory(self):
        with \
                tempfile.TemporaryDirectory() as original_dir, \
                tempfile.TemporaryDirectory() as copy_dir:

            subprocess.run(["git", "init", original_dir])
            subprocess.run(["git", "init", copy_dir])

            uuid_set = UUIDSet(initial_set={
                uuid_0: VersionList(initial_set={
                    "v0": {
                        MetadataPath("path0"): VersionRecord(
                            "1.2",
                            MetadataPath("path0"),
                            create_dataset_tree()
                        )
                    }
                }),
                uuid_1: VersionList(initial_set={
                    "v0.0": {
                        MetadataPath("path0.0"): VersionRecord(
                            "1.3",
                            MetadataPath("path0.0"),
                            create_dataset_tree()
                        )
                    }
                })
            })

            copied_uuid_set = uuid_set.deepcopy(new_destination=copy_dir)
            copied_uuid_set.read_in()
            assert_uuid_sets_equal(self, uuid_set, copied_uuid_set)

    def test_copy_from_backend(self):
        with \
                tempfile.TemporaryDirectory() as original_dir, \
                tempfile.TemporaryDirectory() as copy_dir:

            subprocess.run(["git", "init", original_dir])
            subprocess.run(["git", "init", copy_dir])

            uuid_set = UUIDSet(initial_set={
                uuid_0: VersionList(initial_set={
                    "v0": {
                        MetadataPath("path0"): VersionRecord(
                            "1.2",
                            MetadataPath("path0"),
                            create_dataset_tree()
                        )
                    }
                }),
                uuid_1: VersionList(initial_set={
                    "v0.0": {
                        MetadataPath("path0.0"): VersionRecord(
                            "1.3",
                            MetadataPath("path0.0"),
                            create_dataset_tree()
                        )
                    }
                })
            })

            uuid_set.write_out(original_dir)
            uuid_set_copy = uuid_set.deepcopy(new_destination=copy_dir)

            # assert that the purged uuid-sets are equal
            uuid_set.purge()
            assert_uuid_sets_equal(self, uuid_set, uuid_set_copy)

            # load both and compare the uuid set level
            uuid_set.read_in()
            uuid_set_copy.read_in()
            assert_uuid_sets_equal(self, uuid_set, uuid_set_copy)


if __name__ == '__main__':
    unittest.main()



import os
from typing import Generator, Tuple

from dataladmetadatamodel import JSONObject
from dataladmetadatamodel.filetree import FileTree
from dataladmetadatamodel.metadata import ExtractorConfiguration


DATALAD_DATASET_HIDDEN_DIR_NAME = ".datalad"


def is_dataset_dir(entry: os.DirEntry) -> bool:
    return any(
        filter(
            lambda x: x.is_dir(follow_symlinks=False) and x.name == DATALAD_DATASET_HIDDEN_DIR_NAME,
            os.scandir(entry.path)))


def should_follow(entry: os.DirEntry, ignore_dot_dirs) -> bool:
    return entry.is_dir(follow_symlinks=False) \
           and (not entry.name.startswith(".") or ignore_dot_dirs is False) \
           and not is_dataset_dir(entry)


def read_files(path: str, ignore_dot_dirs: bool = True) -> Generator[Tuple[str, os.DirEntry], None, None]:
    """ Return all sub-entries of prefix_path that are files """

    entries = list(os.scandir(path))
    while entries:
        entry = entries.pop()
        if should_follow(entry, ignore_dot_dirs):
            entries.extend(list(os.scandir(entry.path)))
        else:
            if not entry.is_dir() and not entry.name.startswith("."):
                yield entry.path[len(path) + 1:], entry


def get_extractor_run(path: str,
                      entry: os.DirEntry,
                      parameter_set_count: int) -> JSONObject:

    stat = entry.stat(follow_symlinks=False)
    return {
        "info": f"file-level test metadata for parameter set #{parameter_set_count}",
        "prefix_path": path,
        "size": stat.st_size,
        "atime": stat.st_atime,
        "ctime": stat.st_ctime,
        "mtime": stat.st_mtime
    }


def create_file_tree(mapper_family: str,
                     realm: str,
                     root_dir: str,
                     parameter_set_count: int
                     ) -> FileTree:

    file_tree = FileTree(mapper_family, realm)
    update_file_tree(
        mapper_family,
        realm,
        file_tree,
        root_dir,
        parameter_set_count
    )
    return file_tree


def update_file_tree(mapper_family: str,
                     realm: str,
                     file_tree: FileTree,
                     root_dir: str,
                     parameter_set_count: int):

    for path, entry in read_files(root_dir):
        for count in range(parameter_set_count):
            parameters = {
                "fs_parameter_0": f"value_0.{count}",
                "fs_parameter_1": f"value_1.{count}"
            }
            file_tree.add_extractor_run(
                mapper_family,
                realm,
                path,
                None,
                "file-core-extractor",
                "metadata_creator script",
                "support@datalad.org",
                ExtractorConfiguration(
                    "1.0.0",
                    parameters
                ),
                get_extractor_run(path, entry, count))



import os
import sys
from typing import Optional, Generator, Tuple
from uuid import UUID

from tools.metadata_creator.execute import checked_execute


DATALAD_DATASET_HIDDEN_DIR_NAME = ".datalad"


def get_dataset_id(path) -> Optional[UUID]:
    config_file_path = path + "/.datalad/config"
    try:
        with open(config_file_path) as f:
            for line in f.readlines():
                elements = line.split()
                if elements[:2] == ["id", "="]:
                    return UUID(elements[2])
        print("WARNING: no dataset id in config file: " + config_file_path, file=sys.stderr)
        return None
    except FileNotFoundError:
        print("WARNING: could not open config file: " + config_file_path, file=sys.stderr)
        return None


def has_datalad_dir(path: str) -> bool:
    return any(
        filter(
            lambda e: e.is_dir(follow_symlinks=False) and e.name == DATALAD_DATASET_HIDDEN_DIR_NAME,
            os.scandir(path)))


def is_dataset_dir(entry: os.DirEntry) -> bool:
    return entry.is_dir(follow_symlinks=False) and has_datalad_dir(entry.path)


def should_follow(entry: os.DirEntry, ignore_dot_dirs) -> bool:
    return (
        entry.is_dir(follow_symlinks=False)
        and not entry.name.startswith(".") or ignore_dot_dirs is False)


def get_dataset_version(path) -> Optional[str]:
    git_dir = path + "/.git"
    try:
        return checked_execute(
            ["git", f"--git-dir", git_dir, "log", "-1", "--pretty=format:%H"]
        )[0].strip()
    except RuntimeError:
        return None


def read_datasets(path: str, ignore_dot_dirs: bool = True) -> Generator[Tuple[str, os.DirEntry], None, None]:
    """ Return all datasets and paths """

    path = path.rstrip("/")

    if has_datalad_dir(path):
        path_entry = tuple(filter(lambda e: path.endswith(e.name), os.scandir(path + "/..")))[0]
        yield "", path_entry

    entries = list(os.scandir(path))
    while entries:
        entry = entries.pop()
        if is_dataset_dir(entry):
            yield entry.path[len(path) + 1:], entry
        if should_follow(entry, ignore_dot_dirs):
            entries.extend(list(os.scandir(entry.prefix_path)))


import os
import subprocess
import tempfile
import unittest
from pathlib import Path

from dataladmetadatamodel import version_string
from dataladmetadatamodel.metadatasource import (
    MetadataSource,
    ImmediateMetadataSource,
    LocalGitMetadataSource
)


class TestFactory(unittest.TestCase):
    def test_immediate_factory(self):
        content_obj = {
            "info": "test-content"
        }
        json_obj = {
            "@": {
                "type": "ImmediateMetadataSource",
                "version": version_string
            },
            MetadataSource.TYPE_KEY: ImmediateMetadataSource.TYPE,
            "content": content_obj
        }
        metadata_object = MetadataSource.from_json_obj(json_obj)
        self.assertIsInstance(metadata_object, ImmediateMetadataSource)
        self.assertEqual(metadata_object.content, content_obj)

    def test_local_git_metadata_factory(self):
        repository_path = Path("/a/b/c/1")
        object_reference = "object-reference-1"
        json_obj = {
            "@": {
                "type": "LocalGitMetadataSource",
                "version": version_string
            },
            MetadataSource.TYPE_KEY: LocalGitMetadataSource.TYPE,
            "git_repository_path": repository_path,
            "object_reference": object_reference
        }
        metadata_object = MetadataSource.from_json_obj(json_obj)
        self.assertIsInstance(metadata_object, LocalGitMetadataSource)
        self.assertEqual(metadata_object.git_repository_path, repository_path)
        self.assertEqual(metadata_object.object_reference, object_reference)

    def test_immediate_round_trip(self):
        content = {"test": "test-content-2"}
        immediate_metadata = ImmediateMetadataSource(content)
        created_object = MetadataSource.from_json_obj(
            immediate_metadata.to_json_obj())
        self.assertEqual(immediate_metadata, created_object)

    def test_local_git_metadata_round_trip(self):
        repository_path = Path("/a/b/c/2")
        object_reference = "object-reference-2"
        local_git_metadata = LocalGitMetadataSource(
            repository_path,
            object_reference)
        created_object = MetadataSource.from_json_obj(
            local_git_metadata.to_json_obj())
        self.assertEqual(local_git_metadata, created_object)


class TestFactoryFails(unittest.TestCase):
    def test_missing_type(self):
        json_obj = {
            "@": {
                "type": "ImmediateMetadataSource",
                "version": version_string
            }
        }
        self.assertEqual(MetadataSource.from_json_obj(json_obj), None)

    def test_unknown_type(self):
        json_obj = {
            "@": {
                "type": "ImmediateMetadataSource",
                "version": version_string
            },
            MetadataSource.TYPE_KEY: "....."
        }
        self.assertEqual(MetadataSource.from_json_obj(json_obj), None)

    def test_faulty_local_git_metadata(self):
        json_obj = {
            "@": {
                "type": "LocalGitMetadataSource",
                "version": version_string
            },
            MetadataSource.TYPE_KEY: LocalGitMetadataSource.TYPE
        }
        self.assertEqual(MetadataSource.from_json_obj(json_obj), None)

    def test_faulty_immediate_metadata(self):
        json_obj = {
            "@": {
                "type": "ImmediateMetadataSource",
                "version": version_string
            },
            MetadataSource.TYPE_KEY: ImmediateMetadataSource.TYPE
        }
        self.assertEqual(MetadataSource.from_json_obj(json_obj), None)

    def test_faulty_class_type(self):
        json_obj = {
            "@": {
                "type": ".....",
                "version": version_string
            }
        }
        self.assertEqual(MetadataSource.from_json_obj(json_obj), None)

    def test_faulty_class_version(self):
        json_obj = {
            "@": {
                "type": "LocalGitMetadataSource",
                "version": "..."
            }
        }
        self.assertEqual(MetadataSource.from_json_obj(json_obj), None)


class TestDeepCopy(unittest.TestCase):
    def test_copy_immediate_source(self):
        immediate_metadata = ImmediateMetadataSource("test_content")
        copied_metadata = immediate_metadata.deepcopy()
        self.assertEqual(immediate_metadata, copied_metadata)

    @unittest.skipIf(os.name == "nt", "unix root not handled on windows yet")
    def test_copy_local_git_source(self):
        file_content = "Test file content"
        with \
                tempfile.TemporaryDirectory() as original_dir, \
                tempfile.TemporaryDirectory() as copy_dir:

            subprocess.run(["git", "init", original_dir])
            subprocess.run(["git", "init", copy_dir])

            object_hash = subprocess.check_output(
                f"echo '{file_content}'|git --git-dir {original_dir}/.git "
                f"hash-object --stdin -w",
                shell=True)

            immediate_source = LocalGitMetadataSource(
                Path(original_dir),
                object_hash.decode().strip())

            copied_object_reference = immediate_source.copy_object_to(
                Path(copy_dir))

            self.assertEqual(
                immediate_source.object_reference,
                copied_object_reference)

            copied_file_content = subprocess.check_output(
                f"git --git-dir {copy_dir}/.git cat-file blob "
                f"{copied_object_reference}",
                shell=True)

            self.assertEqual(
                file_content,
                copied_file_content.decode().strip())


if __name__ == '__main__':
    unittest.main()



from uuid import UUID
from typing import (
    Iterable,
    Optional
)
from dataladmetadatamodel.mappableobject import MappableObject
from dataladmetadatamodel.filetree import FileTree
from dataladmetadatamodel.metadata import Metadata
from dataladmetadatamodel.mapper.reference import Reference


class MetadataRootRecord(MappableObject):
    def __init__(self,
                 dataset_identifier: Optional[UUID],
                 dataset_version: Optional[str],
                 dataset_level_metadata: Optional[Metadata],
                 file_tree: Optional[FileTree],
                 realm: Optional[str] = None,
                 reference: Optional[Reference] = None
                 ):

        assert isinstance(dataset_identifier, (type(None), UUID))
        assert isinstance(dataset_version, (type(None), str))
        assert isinstance(dataset_level_metadata, (type(None), Metadata))
        assert isinstance(file_tree, (type(None), FileTree))
        assert isinstance(realm, (type(None), str))
        assert isinstance(reference, (type(None), Reference))

        super().__init__(realm, reference)
        self.dataset_identifier = dataset_identifier
        self.dataset_version = dataset_version
        self.dataset_level_metadata = dataset_level_metadata
        self._file_tree = file_tree

    @staticmethod
    def get_empty_instance(realm: Optional[str] = None,
                           reference: Optional[Reference] = None):
        return MetadataRootRecord(None, None, None, None, realm, reference)

    def modifiable_sub_objects_impl(self) -> Iterable[MappableObject]:
        return [
            child
            for child in [self.dataset_level_metadata, self._file_tree]
            if child is not None
        ]

    def purge_impl(self):
        self.dataset_level_metadata.purge()
        if self._file_tree is not None:
            self._file_tree.purge()
        if self.dataset_level_metadata is not None:
            self.dataset_level_metadata = None
        self._file_tree = None

    def set_file_tree(self, file_tree: FileTree):
        self.ensure_mapped()
        self.touch()
        self._file_tree = file_tree

    def get_file_tree(self):
        self.ensure_mapped()
        if self._file_tree is None:
            return None
        return self._file_tree.read_in()

    file_tree = property(fget=get_file_tree, fset=set_file_tree)

    def set_dataset_level_metadata(self, dataset_level_metadata: Metadata):
        self.ensure_mapped()
        self.touch()
        self.dataset_level_metadata = dataset_level_metadata

    def get_dataset_level_metadata(self):
        self.ensure_mapped()
        return self.dataset_level_metadata.read_in()

    def deepcopy_impl(self,
                      new_mapper_family: Optional[str] = None,
                      new_destination: Optional[str] = None,
                      **kwargs) -> "MetadataRootRecord":

        copied_metadata_root_record = MetadataRootRecord(
            self.dataset_identifier,
            self.dataset_version,
            (
                self.dataset_level_metadata.deepcopy(
                    new_mapper_family,
                    new_destination)
                if self.dataset_level_metadata is not None
                else None
            ),
            (
                self._file_tree.deepcopy(
                    new_mapper_family,
                    new_destination)
                if self._file_tree is not None
                else None
            ),
            None,
            None
        )

        copied_metadata_root_record.write_out(new_destination)
        copied_metadata_root_record.purge()

        return copied_metadata_root_record



import logging

import click
import dataclasses

from tools.metadata_creator.mimic import create_metadata_from_dataset


mdc_logger = logging.getLogger("metadata_creator")


@dataclasses.dataclass
class MDCContext:
    mapper_family: str


@click.group()
@click.pass_context
@click.option("--verbose", is_flag=True, default=False, help="Increase progress output")
@click.option("--quiet", is_flag=True, default=False, help="Do not write progress output")
@click.option("--mapper-family", type=click.Choice(["git", "memory"]), default="git")
def mdc(ctx, verbose, quiet, mapper_family):
    """
    MetaData Creator

    This tool uses the datalad metadata model API and creates
    algorithmically generated metadata structures. This is
    mainly to demonstrate and debug the datalad metadata model
    implementations.

    All metadata is algorithmically created or taken from the
    command arguments. No extractors are executed.

    By default the git-backend is used to persist models, but any
    implemented back can be used. The currently supported backends
    are: `git´ and `memory´.
    """

    if quiet:
        mdc_logger.setLevel(logging.FATAL)
        logging.basicConfig(level=logging.FATAL, format="mdc: %(message)s")

    if verbose:
        mdc_logger.setLevel(logging.DEBUG)
        logging.basicConfig(level=logging.DEBUG, format="mdc: %(message)s")

    ctx.obj = MDCContext(mapper_family=mapper_family)
    mdc_logger.debug(f"context object: {ctx.obj}")


@mdc.command()
@click.pass_context
@click.argument("dataset_path", nargs=1)
@click.argument("realm", nargs=1)
@click.option("-p", "--parameter-set-count", type=int, default=1)
def from_template(ctx, dataset_path, realm, parameter_set_count):
    """
    Create test metadata that mimics the structure of a dataset

    This command will create test metadata the reflects the structure
    of the given dataset. That is, the dataset tree represent the
    dataset containment hierarchy, the UUID set represents the UUIDs
    of the contained datasets and the file trees are modelled after
    the files in the datasets.

    \b
    Usage:
    from-template [DATASET_PATH] [REALM]
    DATASET_PATH: prefix_path of the datalad dataset the should be mimicked
    REALM: realm in which the metadata should be stored
    """

    create_metadata_from_dataset(
        ctx.obj.mapper_family,
        realm,
        dataset_path,
        parameter_set_count)


def main():
    mdc(auto_envvar_prefix="METADATA_CREATOR")


if __name__ == "__main__":
    main()



import logging
import sys
from time import time
from typing import Dict, List, Tuple, Union
from uuid import UUID

from dataladmetadatamodel.datasettree import DatasetTree
from dataladmetadatamodel.filetree import FileTree
from dataladmetadatamodel.metadata import ExtractorConfiguration, Metadata
from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord
from dataladmetadatamodel.uuidset import UUIDSet
from dataladmetadatamodel.versionlist import TreeVersionList, VersionList, VersionRecord


JSONObject = Union[List["JSONObject"], Dict[str, "JSONObject"], int, float, str]

MDC_LOGGER = logging.getLogger("metadata_creator")


FILE_LEVEL_FORMATS = ["file-format-1", "file-format-2"]
DATASET_LEVEL_FORMATS = ["dataset-format-1", "dataset-format-2"]


def _create_tree_paths(tree_spec: List[Tuple[int, int]], upper_levels: List[int]) -> List[str]:
    upper_level_postfix = (
        "." + ".".join(map(str, upper_levels))
        if upper_levels
        else ""
    )
    if len(tree_spec) == 1:
        return [
            f"dataset{upper_level_postfix}.{node_number}"
            for node_number in range(tree_spec[0][0])
        ]

    # If we create a prefix_path that identifies a dataset, add it the the result list
    result = [
        f"dataset{upper_level_postfix}.{node_number}"
        for node_number in range(tree_spec[0][0])
        if node_number < tree_spec[0][1]
    ]

    return result + [
        f"dataset{upper_level_postfix}.{node_number}/" + sub_spec
        if node_number < tree_spec[0][1]
        else f"dir{upper_level_postfix}.{node_number}/" + sub_spec
        for node_number in range(tree_spec[0][0])
        for sub_spec in _create_tree_paths(tree_spec[1:], upper_levels + [node_number])
    ]


def _create_file_paths(tree_spec: List[int], upper_levels: List[int]) -> List[str]:
    upper_level_postfix = (
        "." + ".".join(map(str, upper_levels))
        if upper_levels
        else ""
    )
    node_count = tree_spec[0]
    if len(tree_spec) == 1:
        return [
            f"file{upper_level_postfix}.{node_number}"
            for node_number in range(node_count)]

    return [
        f"dir{upper_level_postfix}.{node_index}/{sub_tree}"
        for node_index in range(node_count)
        for sub_tree in _create_file_paths(tree_spec[1:], upper_levels + [node_index])
    ]


def _create_metadata(mapper_family: str,
                     realm: str,
                     path: str,
                     formats: List[str],
                     parameter_count: int,
                     invocation_count: int) -> Metadata:

    metadata = Metadata(mapper_family, realm)
    for format in formats:
        parameter_lists = [
            {
                f"{format}-parameter-{invocation}.{i}": f"v-{format}.{invocation}.{i}"
                for i in range(parameter_count)
            }
            for invocation in range(invocation_count)
        ]
        for parameter_list in parameter_lists:

            extractor_configuration = ExtractorConfiguration(
                "1.0",
                parameter_list
            )

            metadata.add_extractor_run(
                int(time()),
                format,
                "Auto-created by create.py",
                "christian.moench@web.de",
                extractor_configuration,
                f'{{"type": "inline", "content": '
                f'"{format}({path})"}}'
            )

    return metadata


def _create_file_tree(mapper_family, realm) -> FileTree:

    file_tree = FileTree(mapper_family, realm)
    file_paths = _create_file_paths([3, 4, 10], [])
    for path in file_paths:
        metadata = _create_metadata(
            mapper_family,
            realm,
            path,
            FILE_LEVEL_FORMATS,
            3,
            2
        )
        file_tree.add_metadata(path, metadata)
    return file_tree


def _create_metadata_root_record(mapper_family: str,
                                 realm: str,
                                 path: str,
                                 uuid: UUID,
                                 primary_data_version: str) -> MetadataRootRecord:

    file_tree = _create_file_tree(mapper_family, realm)

    dataset_level_metadata = _create_metadata(
        mapper_family,
        realm,
        path,
        ["dataset_format_1", "dataset_format_2"],
        4,
        2
    )

    metadata_root_record = MetadataRootRecord(
        mapper_family,
        realm,
        uuid,
        primary_data_version,
        Connector.from_object(dataset_level_metadata),
        Connector.from_object(file_tree)
    )
    return metadata_root_record


def _create_dataset_tree(realm) -> DatasetTree:
    dataset_paths = _create_tree_paths([(3, 1), (2, 1), (3, 3)], [])
    dataset_tree = DatasetTree(realm)
    for index, path in enumerate([""] + dataset_paths):
        dataset_tree.add_dataset(
            path,
            _create_metadata_root_record(
                realm,
                path,
                UUID(f"0000000000000000000000000000{index:04x}"),
                "000102030405060708090a0b0c0d0e0f1011{index:04x}".format(index=index)
            )
        )

    return dataset_tree


def main(argv):
    _, mapper_family, realm = argv

    time_counter = 0
    version_counter = 0

    def get_time_str(counter: int) -> str:
        return str(123456789 + counter)

    def get_primary_data_version(counter: int) -> str:
        return "000102030405060708090a0b0c0d0e0f1011{counter:04x}".format(counter=counter)

    dataset_tree = _create_dataset_tree(mapper_family, realm)
    dataset_tree_version_list = TreeVersionList(initial_set={
        dataset_tree.value.dataset_version: VersionRecord(
            get_time_str(time_counter),
            None,
            Connector.from_object(dataset_tree)
        )
    })

    # Extract the UUID Set from the dataset tree:
    uuid_version_lists = {}
    for path, metadata_root_record in dataset_tree.dataset_paths:
        uuid_version_lists[metadata_root_record.dataset_identifier] = Connector.from_object(
            VersionList(initial_set={
                get_primary_data_version(version_counter): VersionRecord(
                    get_time_str(time_counter),
                    path,
                    Connector.from_object(metadata_root_record)
                )
            })
        )
        version_counter += 1
        time_counter += 1

    uuid_set = UUIDSet(
        mapper_family,
        realm,
        uuid_version_lists
    )

    print(uuid_set.save())
    print(dataset_tree_version_list.save())


if __name__ == "__main__":
    main(sys.argv)



import unittest
from unittest.mock import patch

from ..reference import Reference


class TestReference(unittest.TestCase):

    def check_json_values(self, json_object):
        for value in json_object.values():
            if isinstance(value, (dict, list)):
                self.check_json_values(value)
            else:
                self.assertIsInstance(value, (str, int, float))

    def test_json_serialisation(self):
        reference = Reference("SomeClass", "some-location")
        json_object = reference.to_json_obj()
        self.check_json_values(json_object)
        self.assertEqual(json_object["class_name"], "SomeClass")
        self.assertEqual(json_object["location"], "some-location")

    def test_json_serialisation_deserialization(self):
        reference1 = Reference("SomeClass", "some-location")
        json_object = reference1.to_json_obj()
        reference2 = Reference.from_json_obj(json_object)
        self.assertEqual(reference1, reference2)

    def test_old_style_warning(self):
        reference1 = Reference("SomeClass", "some-location")

        json_object = reference1.to_json_obj()
        json_object["mapper_family"] = "git"
        json_object["realm"] = "/tmp/t"

        with patch("dataladmetadatamodel.mapper.reference.logger") as logger:
            reference2 = Reference.from_json_obj(json_object)
            self.assertEqual(reference1, reference2)
            logger.info.assert_called_once()

    def test_remote(self):
        for realm in ("git:x.com", "http://x.com/", "https://x.com/", "ssh:x@x.com"):
            self.assertTrue(Reference.is_remote(realm))
            self.assertFalse(Reference.is_local(realm))

    def test_local(self):
        for realm in ("a/b/c", "C:\\x", "c:\\x", "/a/b/c"):
            self.assertFalse(Reference.is_remote(realm))
            self.assertTrue(Reference.is_local(realm))


if __name__ == '__main__':
    unittest.main()



import json
import shlex
import subprocess
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union
)

from dataladmetadatamodel.log import logger
from dataladmetadatamodel.mapper.reference import Reference


def execute(arguments: Union[str, List[str]],
            stdin_content: Optional[Union[str, bytes]] = None) -> Any:

    logger.debug(f"gitbackend: execute({arguments}, {stdin_content})")
    return subprocess.run(
        shlex.split(arguments) if isinstance(arguments, str) else arguments,
        input=(
            stdin_content.encode()
            if isinstance(stdin_content, str)
            else stdin_content),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)


def checked_execute(arguments: Union[str, List[str]],
                    stdin_content: Optional[Union[str, bytes]] = None
                    ) -> Tuple[List[str], List[str]]:

    result = execute(arguments, stdin_content)
    if result.returncode != 0:
        error = RuntimeError(
            f"Command failed (exit code: {result.returncode}) "
            f"{' '.join(arguments)}:\n"
            f"STDOUT:\n"
            f"{result.stdout.decode()}"
            f"STDERR:\n"
            f"{result.stderr.decode()}")
        error.stdout = result.stdout.decode()
        error.stderr = result.stderr.decode()
        raise error
    return (
        result.stdout.decode().splitlines(),
        result.stderr.decode().splitlines())


def git_command_line(repo_dir: str,
                     command: str,
                     arguments: List[str]
                     ) -> List[str]:

    return [
               "git",
               "-P",
               "--git-dir",
               repo_dir + "/.git",
               command
           ] + arguments


def git_text_result(cmd_line: Union[str, List[str]]):
    result = checked_execute(cmd_line)[0]
    return "\n".join(result)


def adapt_for_remote(repo_dir: str, object_reference: str) -> str:
    from dataladmetadatamodel.mapper.gitmapper.localcache import (
        cache_object,
        get_cache_realm,
    )

    if Reference.is_remote(repo_dir):
        cache_object(repo_dir, object_reference)
        return str(get_cache_realm(repo_dir))
    return repo_dir


def git_load_str(repo_dir: str, object_reference: str) -> str:
    repo_dir = adapt_for_remote(repo_dir, object_reference)
    cmd_line = git_command_line(repo_dir, "show", [object_reference])
    return git_text_result(cmd_line)


def git_load_json(repo_dir: str, object_reference: str) -> Union[Dict, List]:
    repo_dir = adapt_for_remote(repo_dir, object_reference)
    return json.loads(git_load_str(repo_dir, object_reference))


def git_init(repo_dir: str) -> None:
    cmd_line = ["git", "init", repo_dir]
    checked_execute(cmd_line)


def git_object_exists_locally(repo_dir: str,
                              object_reference: str) -> bool:
    cmd_line = git_command_line(repo_dir, "cat-file", ["-e", object_reference])
    return execute(cmd_line).returncode == 0


def git_read_tree_node(repo_dir: str,
                       object_reference: str) -> List[str]:
    repo_dir = adapt_for_remote(repo_dir, object_reference)
    cmd_line = git_command_line(repo_dir, "cat-file", ["-p", object_reference])
    return checked_execute(cmd_line)[0]


def git_ls_tree(repo_dir: str, object_reference: str) -> List[str]:
    repo_dir = adapt_for_remote(repo_dir, object_reference)
    cmd_line = git_command_line(repo_dir, "ls-tree", [object_reference])
    return checked_execute(cmd_line)[0]


def git_ls_tree_recursive(repo_dir: str,
                          object_reference: str,
                          show_intermediate: bool = False) -> List[str]:
    repo_dir = adapt_for_remote(repo_dir, object_reference)
    if show_intermediate is True:
        cmd_line = git_command_line(repo_dir, "ls-tree", ["-r", "-t", object_reference])
    else:
        cmd_line = git_command_line(repo_dir, "ls-tree", ["-r", object_reference])
    return checked_execute(cmd_line)[0]


def git_save_str(repo_dir: str, content: str) -> str:
    cmd_line = git_command_line(repo_dir, "hash-object", ["-w", "--stdin"])
    return checked_execute(cmd_line, stdin_content=content)[0][0]


def git_save_file_list(repo_dir: str, file_list: List[str]) -> List[str]:
    cmd_line = git_command_line(
        repo_dir,
        "hash-object",
        ["-w", "--no-filters", "--stdin-paths"])
    return checked_execute(
        cmd_line,
        stdin_content="\n".join(file_list))[0]


def git_save_json(repo_dir: str, json_object: Union[Dict, List]) -> str:
    return git_save_str(repo_dir, json.dumps(json_object))


def git_save_tree_node(repo_dir: str,
                       entry_set: Iterable[Tuple[str, str, str, str]]
                       ) -> str:

    tree_spec = "\n".join([
        f"{flag} {node_type} {object_hash}\t{name}"
        for flag, node_type, object_hash, name in entry_set
    ]) + "\n"
    cmd_line = git_command_line(repo_dir, "mktree", ["--missing", ])
    return checked_execute(cmd_line, stdin_content=tree_spec)[0][0]


def git_save_tree(repo_dir: str,
                  entry_set: Iterable[Tuple[str, str, str, str]]
                  ) -> str:
    tree_spec = "\n".join([
        f"{flag} {node_type} {object_hash}\t{name}"
        for flag, node_type, object_hash, name in entry_set
    ]) + "\n"
    cmd_line = git_command_line(repo_dir, "mktree", ["--missing", ])
    return checked_execute(cmd_line, stdin_content=tree_spec)[0][0]


def git_update_ref(repo_dir: str, ref_name: str, location: str) -> None:
    cmd_line = git_command_line(
        repo_dir,
        "update-ref",
        [ref_name, location])
    checked_execute(cmd_line)


def git_delete_ref(repo_dir: str, ref_name: str) -> None:
    cmd_line = git_command_line(
        repo_dir,
        "update-ref",
        ["-d", ref_name])
    checked_execute(cmd_line)


def git_fetch_object(repo_dir: str,
                     remote_repo: str,
                     remote_reference: str):
    return git_fetch_reference(repo_dir, remote_repo, remote_reference)


def git_fetch_reference(repo_dir: str,
                        remote_repo: str,
                        remote_reference: str,
                        local_reference: Optional[str] = None):
    cmd_line = git_command_line(
        repo_dir,
        "fetch",
        [
            remote_repo,
            "-f",
            "--no-tags",
            "--no-recurse-submodules",
            remote_reference + (
                f":{local_reference}"
                if local_reference is not None
                else ""
            )
        ]
    )
    checked_execute(cmd_line)



from collections import defaultdict
from typing import (
    DefaultDict,
    Dict
)

from .gitmapper import GIT_MAPPER_LOCATIONS


git_mapper_family_name = "git"

locations = {
    git_mapper_family_name: GIT_MAPPER_LOCATIONS
}

registered_mapper: DefaultDict[str, Dict[str, "Mapper"]] = defaultdict(dict)


def _get_locations(mapper_family: str):
    return locations.get(mapper_family, (None, None))


def get_uuid_set_location(mapper_family: str):
    return _get_locations(mapper_family)[1]


def get_tree_version_list_location(mapper_family: str):
    return _get_locations(mapper_family)[0]


def get_mapper(class_name: str, backend_type: str) -> "Mapper":
    return registered_mapper[backend_type][class_name]


def set_mapper(class_name: str, backend_type: str, mapper: "Mapper"):
    registered_mapper[backend_type][class_name] = mapper


def initialize_object_store():
    from .gitmapper.metadatamapper import MetadataGitMapper
    from .gitmapper.metadatarootrecordmapper import MetadataRootRecordGitMapper
    from .gitmapper.mtreenodemapper import MTreeNodeGitMapper
    from .gitmapper.textmapper import TextGitMapper
    from .gitmapper.uuidsetmapper import UUIDSetGitMapper
    from .gitmapper.versionlistmapper import (
        TreeVersionListGitMapper,
        VersionListGitMapper
    )

    set_mapper("DatasetTree", "git", MTreeNodeGitMapper("MTreeNode"))
    set_mapper("FileTree", "git", MTreeNodeGitMapper("MTreeNode"))
    set_mapper("Metadata", "git", MetadataGitMapper("Metadata"))
    set_mapper("MetadataRootRecord", "git", MetadataRootRecordGitMapper("MetadataRootRecord"))
    set_mapper("MTreeNode", "git", MTreeNodeGitMapper("MTreeNode"))
    set_mapper("Text", "git", TextGitMapper("Text"))
    set_mapper("TreeVersionList", "git", TreeVersionListGitMapper("TreeVersionList"))
    set_mapper("UUIDSet", "git", UUIDSetGitMapper("UUIDSet"))
    set_mapper("VersionList", "git", VersionListGitMapper("VersionList"))


initialize_object_store()
